#include <gazebo/gazebo.hh>
#include <sdf/sdf.hh>
#include <gazebo/common/Plugin.hh>
#include <gazebo/msgs/msgs.hh>
#include <gazebo/physics/physics.hh>
#include <gazebo/transport/transport.hh>
#include <octomap_msgs/Octomap.h>
#include <octomap_msgs/conversions.h>
#include <octomap/octomap.h>
#include <gazebo/transport/transport.hh>
#include <gazebo/util/system.hh>
#include <ros/ros.h>
#include <stdio.h>
#include <stdlib.h>
#include <std_msgs/String.h>
#include "build/msgs/ustring.pb.h"
#include "boost/thread/mutex.hpp"
#include <ros/callback_queue.h>
#include <ros/subscribe_options.h>
#include <fstream>
#include <unistd.h>
#include <boost/format.hpp>
#include <time.h>

// function to compare doubles and floats
// arguments are the two double/floats to compare, and the epsilon value
// which signifies the percision of the comparision
bool double_equals(float a, float b, float epsilon = 0.00001)
{
  return std::fabs(a - b) < epsilon;
}

class Voxel {
  private:
  	octomap::point3d center; // the center point of the voxel
  	octomap::point3d dimensions; // the dimensions of the voxel
  public:
    // Voxel class constructor, arguments are the center in a point3d, and the x, y, and z dimensions as floats
  	Voxel(octomap::point3d _center, float size, float length, float height): center(_center), dimensions(size, length, height) {}

    // Voxel class constructor, arguments are the center point and dimenions, both as a point3d
    Voxel(octomap::point3d _center, octomap::point3d _dimensions): center(_center), dimensions(_dimensions) {}

    // Getter function that returns center point of the voxel
  	octomap::point3d getCenter() const {
  		return center;
  	}

    // Getter function that returns dimensions of the voxel
  	octomap::point3d getDimensions() const {
  		return dimensions;
  	}

    // Check if this voxel is adjacent to another and lie on the same axis in one dimension
    // arguments are a voxel, and an int to specify the axis dimension
    // 1 for z, 2 for y, 3 for x.
    // This int corresponds with the mode of the VoxelCollection class
  	bool adjacent(const Voxel voxel, int dimension) const {
  		if(dimension == 1) {
  			return double_equals(center.x(), voxel.getCenter().x()) && double_equals(center.y(), voxel.center.y()) && 
  			double_equals(center.distance(voxel.getCenter()), (dimensions.z() + voxel.getDimensions().z())/2.0);
  		}
  		else if (dimension == 2) {
  			return double_equals(center.x(), voxel.getCenter().x()) && double_equals(center.z(), voxel.center.z()) && 
  			double_equals(center.distance(voxel.getCenter()), (dimensions.y() + voxel.getDimensions().y())/2.0);
  		}
  		else {
  			return double_equals(center.y(), voxel.getCenter().y()) && double_equals(center.z(), voxel.center.z()) && 
  			double_equals(center.distance(voxel.getCenter()), (dimensions.x() + voxel.getDimensions().x())/2.0);
  		}
  	}

    // Check if this voxel two dimensions of the same size with another one
    // arguments are a voxel, and an int to specify the dimensions
    // 1 for x and y, 2 for x and z, 3 for y and z.
    // This int corresponds with the mode of the VoxelCollection class
  	bool equal_cross_section(const Voxel voxel, int dimension) const {
  		if(dimension == 1) {
  			return double_equals(dimensions.x(), voxel.getDimensions().x()) && double_equals(dimensions.y(), voxel.getDimensions().y());
  		}
  		else if(dimension == 2) {
  			return double_equals(dimensions.x(), voxel.getDimensions().x()) && double_equals(dimensions.z(), voxel.getDimensions().z());
  		}
  		else {
  			return double_equals(dimensions.y(), voxel.getDimensions().y()) && double_equals(dimensions.z(), voxel.getDimensions().z());
  		}
  	}
};

// Comparetor function to sort voxels on their center z value
bool voxelZComparator(Voxel i, Voxel j) {
  return i.getCenter().z() < j.getCenter().z();
}

// Comparetor function to sort voxels on their center y value
bool voxelYComparator(Voxel i, Voxel j) {
  return i.getCenter().y() < j.getCenter().y();
}

// Comparetor function to sort voxels on their center x value
bool voxelXComparator(Voxel i, Voxel j) {
  return i.getCenter().x() < j.getCenter().x();
}

// helper function to print out the center point and dimensions of a voxel
void printVoxel(const Voxel voxel) {
  octomap::point3d centerpoint = voxel.getCenter();
  octomap::point3d dims = voxel.getDimensions();
  printf("Voxel x: %f y: %f z: %f size: %f length: %f height: %f\n", centerpoint.x(), centerpoint.y(), centerpoint.z(), dims.x(), dims.y(), dims.z());
}

// Class to hold a collection of voxels
// This class is designed to merge groups of adjacent, same sized, voxels along an axis (x, y, or z)
// The mode member of this class determines the axis
// 1 is z, 2 is y, 3 is x
class VoxelCollection {
  private:
  	std::vector<std::vector<Voxel > > voxels; //store all the voxels in vectors, with voxels that can potentially be merged together being stored in their own vector
  	int mode; //an integer the signify the axis that will be merged on (1 for z, 2 for y, 3 for x)

    // function to return a single Voxel from a vector of adjacent voxels
    // arguments are a vector of adjacent voxels
    Voxel mergeVoxels(const std::vector<Voxel> to_merge) const {
      Voxel merged_voxel = to_merge[0];
      octomap::point3d old_center = to_merge[0].getCenter();
      octomap::point3d old_dims = to_merge[0].getDimensions();
      if(mode == 1) {
        octomap::point3d new_center(old_center.x(), old_center.y(), (old_center.z() + to_merge.back().getCenter().z())/2.0);
        octomap::point3d new_dims(old_dims.x(), old_dims.y(), (old_dims.z() * to_merge.size()));
        merged_voxel = Voxel(new_center, new_dims);
      }
      else if(mode == 2) {
        octomap::point3d new_center(old_center.x(), (old_center.y() + to_merge.back().getCenter().y())/2.0, old_center.z());
        octomap::point3d new_dims(old_dims.x(), (old_dims.y() * to_merge.size()), old_dims.z());
        merged_voxel = Voxel(new_center, new_dims);
      }
      else {
        octomap::point3d new_center((old_center.x() + to_merge.back().getCenter().x())/2.0, old_center.y(), old_center.z());
        octomap::point3d new_dims((old_dims.x() * to_merge.size()), old_dims.y(), old_dims.z());
        merged_voxel = Voxel(new_center, new_dims);
      }
      return merged_voxel;
    }

  public:
    // Voxel Collection class constuctor. Arguments are an int for the mode (1 for z, 2 for y, 3 for x)
    VoxelCollection(int _mode): mode(_mode) {}

    // Add a voxel to the voxel collection
    // stores the voxel in the vector of voxel vector voxels, putting it in an existing voxel vector if one exists that
    // could have voxels to with merge with the new one, and if not making a new vector for the voxel
    // arguments are the voxel to add
    void add(Voxel voxel) {
      int foundindex = -1; //if a good existing vector is found, it's index will be stored in this variable
      static int voxel_vector_count = 0;
      // for each existing vector of Voxel in voxels, check and see if the new voxel could be added to it, if one is found, break
      for (int i = 0; i < voxels.size(); i++) {
        octomap::point3d centerpoint = voxels[i][0].getCenter();
        switch(mode) {
          case 1:
            foundindex = double_equals(centerpoint.x(),voxel.getCenter().x()) && double_equals(centerpoint.y(),voxel.getCenter().y()) ? i : -1;
            break;
          case 2:
            foundindex = double_equals(centerpoint.x(),voxel.getCenter().x()) && double_equals(centerpoint.z(),voxel.getCenter().z()) ? i : -1;
            break;
          case 3:
            foundindex = double_equals(centerpoint.y(),voxel.getCenter().y()) && double_equals(centerpoint.z(),voxel.getCenter().z()) ? i : -1;
            break;
        }
        if(foundindex != -1) {
          break;
        }
      }
      // if a good existing vector was found, add the new Voxel to it
      if (foundindex != -1) {
        //printf("Putting into existing voxel vector\n");
        voxels[foundindex].push_back(voxel);
      }
      // if not, create a new vector for the Voxel
      else {
        //printf("Making new voxel vector, column %d, x=%f, y=%f\n", voxel_vector_count, voxel.getCenter().x(), voxel.getCenter().y());
        std::vector<Voxel> new_voxel_vector;
        new_voxel_vector.push_back(voxel);
        voxels.push_back(new_voxel_vector);
        voxel_vector_count++;
      }
    }

    // Merge all of the adjacent voxels in each vector of Voxel into one Voxel
    // This function should only be called once everything has been added to the VoxelCollection
    void merge() {
      //for each vector of Voxel in voxels, sort the vector based on the mode of the VoxelCollection, then merge adjacent voxels in that vector
      for(int i = 0; i < voxels.size(); i++) {
        //printf("Merging column %d of size %lu\n", i, voxels[i].size());
        if(mode == 1) {
          std::sort(voxels[i].begin(), voxels[i].end(), voxelZComparator);
        }
        else if(mode == 2) {
          std::sort(voxels[i].begin(), voxels[i].end(), voxelYComparator);
        }
        else {
          std::sort(voxels[i].begin(), voxels[i].end(), voxelXComparator);
        }
        std::vector<Voxel> temp; // vector to hold merged voxels (single voxels created from a vector of voxels)
        std::vector<Voxel> consecutive_voxels; // vector to hold consecutive voxels
        consecutive_voxels.push_back(voxels[i][0]); // put the first voxel in the vector of consecutive voxels to get things started
        int times_merged = 0;
        // for all the voxels in the sorted vector, put them in a vector if they are adjacent
        // once an unadjacent voxel is found, merge the vector of adjacent voxels, store the single merged voxel, and place the non-adjacent one into the vector
        // after the loop, there are still voxels in consective voxels, so merge them together and store them
        for(int j = 1; j < voxels[i].size(); j++) {
          if(consecutive_voxels.back().adjacent(voxels[i][j], mode) && consecutive_voxels.back().equal_cross_section(voxels[i][j], mode)) {
            consecutive_voxels.push_back(voxels[i][j]);
          }
          else {
            temp.push_back(mergeVoxels(consecutive_voxels));
            times_merged++;
            consecutive_voxels.clear();
            consecutive_voxels.push_back(voxels[i][j]);
          }
        }
        temp.push_back(mergeVoxels(consecutive_voxels));
        times_merged++;
        //printf("made %d merges\n", times_merged);
        voxels[i].swap(temp); //update the vector to have the merged voxels
      }
    }

    // return of vector of all Voxel in voxels
    std::vector<Voxel> getVoxels() {
      std::vector<Voxel> v;
      for(int i = 0; i < voxels.size(); i++) {
        for(int j = 0; j < voxels[i].size(); j++) {
          v.push_back(voxels[i][j]);
        }
      }
      return v;
    }
};

// function to test if merging works
void testVoxels() {
  Voxel voxel1(octomap::point3d(0.0, 0.0, 0.0), 1.0, 1.0, 1.0);
  Voxel voxel2(octomap::point3d(0.0, 0.0, 1.0), 1.0, 1.0, 1.0);
  Voxel voxel3(octomap::point3d(0.0, 0.0, 2.0), 1.0, 1.0, 1.0);
  Voxel voxel4(octomap::point3d(0.0, 1.0, 0.0), 1.0, 1.0, 1.0);
  Voxel voxel5(octomap::point3d(0.0, 2.0, 0.0), 1.0, 1.0, 1.0);
  Voxel voxel6(octomap::point3d(1.0, 2.0, 0.0), 1.0, 1.0, 1.0);
  Voxel voxel7(octomap::point3d(1.0, 2.0, -1.0), 1.0, 1.0, 1.0);
  Voxel voxel8(octomap::point3d(1.0, 2.0, 1.0), 1.0, 1.0, 1.0);
  Voxel voxel9(octomap::point3d(1.0, 2.0, 2.0), 1.0, 1.0, 1.0);
  Voxel voxel10(octomap::point3d(0.0, 2.0, 1.0), 1.0, 1.0, 1.0);
  Voxel voxel11(octomap::point3d(0.0, 2.0, 2.0), 1.0, 1.0, 1.0);
  Voxel voxel12(octomap::point3d(0.0, 2.0, 4.0), 1.0, 1.0, 1.0);
  Voxel voxel13(octomap::point3d(0.0, 3.0, -1.0), 1.0, 1.0, 1.0);
  Voxel voxel14(octomap::point3d(1.0, 2.0, -5.0), 1.0, 1.0, 1.0);
  Voxel voxel15(octomap::point3d(1.0, 2.0, 3.0), 1.0, 1.0, 1.0);
  Voxel voxel16(octomap::point3d(1.0, 2.0, 4.0), 1.0, 1.0, 1.0);
  Voxel voxel17(octomap::point3d(0.0, 3.0, 0.0), 1.0, 1.0, 1.0);
  Voxel voxel18(octomap::point3d(0.0, 3.0, 1.0), 1.0, 1.0, 1.0);
  Voxel voxel19(octomap::point3d(0.0, 2.0, 3.0), 1.0, 1.0, 1.0);
  VoxelCollection voxel_xy(1);
  voxel_xy.add(voxel1);
  voxel_xy.add(voxel2);
  voxel_xy.add(voxel3);
  voxel_xy.add(voxel4);
  voxel_xy.add(voxel6);
  voxel_xy.add(voxel7);
  voxel_xy.add(voxel8);
  voxel_xy.add(voxel9);
  voxel_xy.add(voxel10);
  voxel_xy.add(voxel11);
  voxel_xy.add(voxel12);
  voxel_xy.add(voxel13);
  voxel_xy.add(voxel14);
  voxel_xy.add(voxel15);
  voxel_xy.add(voxel16);
  voxel_xy.add(voxel17);
  voxel_xy.add(voxel18);
  voxel_xy.add(voxel19);
  voxel_xy.merge();
} 

using namespace gazebo;
// Class for the world plugin in Gazebo
// responsible for recieving the octomap, using Voxel and VoxelCollection to reduce the number of voxels, and then rendering them into Gazebo
class ubelluris_gazebo_plugin : public WorldPlugin {
  private:
    transport::NodePtr node; // the node to represent the plugin within the greater Gazebo system
    physics::WorldPtr world; // a pointer to the Gazebo world
    sdf::ElementPtr sdf; // a pointer to Gazebo's sdf
    ros::NodeHandle* rosnode_; // the node to represent the plugin within the greater ROS system
    ros::CallbackQueue queue_; // a queue to hold all requests for the callback to render an octomap
    boost::thread callback_queue_thread_; // a thread for the requests to render an octomap
    ros::Subscriber sub; // a subscriber to subscribe to the /ubelluris/octomap_msgs topic
    double x_offset; // the ofset in the x direction for voxels to be rendered in
    double y_offset; // the ofset in the y direction for voxels to be rendered in
    double z_offset; // the ofset in the z direction for voxels to be rendered in

  public:

    // class destructor
    ~ubelluris_gazebo_plugin() {
      this->queue_.clear();
      this->queue_.disable();
      this->rosnode_->shutdown();
      this->callback_queue_thread_.join();
      delete this->rosnode_;
    }

    // callback function to render an octomap into the world
    void CallBack(const std_msgs::String::ConstPtr& msg) {
      // assign x, y, and z offsets based on the position of the atlas robot's pelvis
      time_t start_time = time(NULL);
      physics::ModelPtr atlasModel = this->world->GetModel("atlas");
      physics::LinkPtr pelvisLink = atlasModel->GetLink("atlas::pelvis");
      physics::LinkState * pelvisState = new physics::LinkState(pelvisLink);
      math::Pose pose = pelvisState->GetPose();
      this->x_offset = pose.pos.x;
      this->y_offset = pose.pos.y;
      this->z_offset = pose.pos.z;
      this->world->SetPaused(true);
      printf("x offset: %f, y offset: %f, z offset: %f\n", this->x_offset, this->y_offset, this->z_offset);
      char path[MAXPATHLEN];
      getcwd(path, MAXPATHLEN);
      // get the octomap from the message and expand it
      octomap::OcTree * octree = new octomap::OcTree(msg->data);
      octree->expand();
      // render the octomap into the world
      spawn_models_from_octomap(octree);
      delete(octree);
      this->world->SetPaused(false);
      time_t end_time = time(NULL);
      double run_time = difftime(end_time, start_time);
      printf("completed in %f seconds\n", run_time);
    }

  // render and octomap into the world. arguments are the octomap
  void spawn_models_from_octomap(const octomap::OcTree * p_octtree) {
    std::vector<Voxel> initial_voxels;
    int base_voxel_count = 0;
    // for each node in the octomap, if it is occupied, make it into a Voxel and then add it to initial_voxels
    for(octomap::OcTree::leaf_iterator it = p_octtree->begin_leafs(16); it != p_octtree->end_leafs(); ++it) {
      if(it->getOccupancy() > 0.75) {
        Voxel vox(it.getCoordinate(), it.getSize(), it.getSize(), it.getSize());
        initial_voxels.push_back(vox);
        base_voxel_count++;
      }
    }
    printf("started with %d voxels\n", base_voxel_count);
    // create a vector of all the different merge orders
    std::vector< boost::tuples::tuple<int, int, int> > merges;
    int best_merge_index = 0;
    merges.push_back(boost::tuples::tuple<int, int, int>(1, 2, 3));
    merges.push_back(boost::tuples::tuple<int, int, int>(1, 3, 2));
    merges.push_back(boost::tuples::tuple<int, int, int>(2, 1, 3));
    merges.push_back(boost::tuples::tuple<int, int, int>(2, 3, 1));
    merges.push_back(boost::tuples::tuple<int, int, int>(3, 2, 1));
    merges.push_back(boost::tuples::tuple<int, int, int>(3, 1, 2));
    // set the inital best_final_voxels to be the results from merging on the values from the first item in the vector
    std::vector<Voxel> best_final_voxels = mergeVoxels(initial_voxels, merges[0].get<0>(), merges[0].get<1>(), merges[0].get<2>());
    // calculate the voxels for each other merge order in the vector. If a better one is found, use it instead
    for(int i = 1; i < merges.size(); i++) {
      std::vector<Voxel> next_merge = mergeVoxels(initial_voxels, merges[i].get<0>(), merges[i].get<1>(), merges[i].get<2>());
      printf("Merge %d is of size %lu\n", i, next_merge.size());
      if(next_merge.size() < best_final_voxels.size()) {
        best_final_voxels = next_merge;
        best_merge_index = i;
      }
    }
    int object_count = best_final_voxels.size();
    // for each of the voxels, render them into the world
    for(int i = 0; i < best_final_voxels.size(); i++) {
      sdf::SDF boxSDF;
      std::string sdfstr = makeBoxSdf(best_final_voxels[i].getCenter().x(), best_final_voxels[i].getCenter().y(), best_final_voxels[i].getCenter().z(),
        best_final_voxels[i].getDimensions().x(), best_final_voxels[i].getDimensions().y(), best_final_voxels[i].getDimensions().z());
      boxSDF.SetFromString(sdfstr);
      //printf("Drawing object %d\n", object_count);
      this->world->InsertModelSDF(boxSDF);
    }
    printf("Went from %d voxels to %d objects, best merge was %d\n", base_voxel_count, object_count, best_merge_index);
  }

  std::vector<Voxel> mergeVoxels(std::vector<Voxel> voxels, int first_merge, int second_merge, int third_merge) {
    printf("Merging %d %d %d\n", first_merge, second_merge, third_merge);
    VoxelCollection voxels_one(first_merge); // VoxelCollection that will merge on the first dimension
    int base_voxel_count = voxels.size();
    // for each voxel, add it to voxels_one
    for(int i = 0; i < voxels.size(); i++) {
      voxels_one.add(voxels[i]);
    }
    printf("started with %d voxels\n", base_voxel_count);
    voxels_one.merge(); // merge all the voxels now that they have been added
    // now we merge on the second dimension
    std::vector<Voxel> merged_voxels_one = voxels_one.getVoxels(); // get all the (now merged) voxels from voxels_one
    int column_count = merged_voxels_one.size();
    VoxelCollection voxels_two(second_merge); // VoxelCollection that will merge on the second dimension
    // add all of the voxels that went through the first merge into voxels_two
    for(int i = 0; i < merged_voxels_one.size(); i++) {
      voxels_two.add(merged_voxels_one[i]);
    }
    printf("Went from %d voxels to %d columns\n", base_voxel_count, column_count);
    voxels_two.merge(); // merge all the voxels now that they have been added
    std::vector<Voxel> merged_voxels_two = voxels_two.getVoxels();
    // finally, we merge on the third dimension
    int plane_count = merged_voxels_two.size();
    VoxelCollection voxels_three(third_merge); // VoxelCollection that will merge on the third dimension
    // add all the voxels that went through the second merge into voxels_three
    for(int i = 0; i < merged_voxels_two.size(); i++) {
      voxels_three.add(merged_voxels_two[i]);
    }
    printf("Went from %d columns to %d planes\n", column_count, plane_count);
    voxels_three.merge(); // merge all the voxels now that they have been added
    return voxels_three.getVoxels(); // return the final results of the Voxel merging
  }

  // return a string for the SDF of a box to be rendered into the world
  // arguments are the x, y, and z values of the center point of the box, and the size of the x, y, and z dimensions of the box, as floats
  std::string makeBoxSdf(float x, float y, float z, float width, float length, float height) {
    static int model_count = 0;
    char sizeboxstr[128];
    sprintf(sizeboxstr,"<box><size> %f %f %f </size></box>",width,length,height);
    std::ostringstream newModelStr;
    newModelStr << "<sdf version ='1.5'>"
    << "<model name='ubelluris_voxel_" << model_count <<"'>"
    << "<pose> " << x + this->x_offset << " " << y + this->y_offset << " " << z + this->z_offset <<" 0 0 0</pose>"
    << "<static>true</static>"
    << "<link name ='link'>"
    <<   "<inertial><mass>1.0</mass></inertial>"
    <<   "<collision name ='collision'>"
    <<     "<geometry>"
    <<       sizeboxstr
    <<     "</geometry>"
    << "</collision>"
    << "<visual name ='visual'>"
    <<     "<geometry>"
    <<    sizeboxstr
    <<     "</geometry>"
    <<     "<material>"
    <<       "<script>"
    <<         "<uri>file://media/materials/scripts/gazebo.material</uri>"
    <<         "<name>Gazebo/Grey</name>"
    <<       "</script>"
    <<     "</material>"
    <<   "</visual>"
    << "</link>"
    << "</model>"
    << "</sdf>";
    model_count++;
    return newModelStr.str();
  }

  // Class constructor
  ubelluris_gazebo_plugin() {
    this->rosnode_ = NULL;
    printf("UBELLURIS GAZEBO PLUGIN LOADED\n");
  }

  // the Load function is called when the plugin's node is first initalized
  // this function sets up the node so that it is subscribe to the correct topic
  void Load(physics::WorldPtr _world, sdf::ElementPtr _sdf) {
    printf("hello world\n");
    GZ_ASSERT(_world, "WORLD PTR IS NULL!");
    GZ_ASSERT(_sdf, "SDF PTR IS NULL");
    this->world = _world; 
    this->node = transport::NodePtr(new transport::Node());
    this->node->Init(_world->GetName());
    this->sdf = _sdf;
    printf("after subscribed\n");
    testVoxels();
    // initialize ros
    if (!ros::isInitialized()) {
      int argc = 0;
      char** argv = NULL;
      ros::init(argc,argv,"gazebo",ros::init_options::NoSigintHandler|ros::init_options::AnonymousName);
    }
    this->rosnode_ = new ros::NodeHandle();
    // Custom Callback Queue
    ros::SubscribeOptions so2 = ros::SubscribeOptions::create<std_msgs::String>(std::string("/ubelluris/octomap_msg"),1,boost::bind( &ubelluris_gazebo_plugin::CallBack,this,_1),ros::VoidPtr(), &this->queue_);
    this->sub = this->rosnode_->subscribe(so2);
    this->callback_queue_thread_ = boost::thread( boost::bind( &ubelluris_gazebo_plugin::QueueThread,this ) );
  }

  // function for the subscriber queue thread
  void QueueThread() {
    static const double timeout = 0.01;
    while (this->rosnode_->ok()) {
      this->queue_.callAvailable(ros::WallDuration(timeout));
    }
  }

};

  GZ_REGISTER_WORLD_PLUGIN(ubelluris_gazebo_plugin);
